/*
 * CallWeaver -- An open source telephony toolkit.
 *
 * Copyright (c) 2004 - 2005, inAccess Networks
 *
 * Michael Manousos <manousos@inaccessnetworks.com>
 *
 * See http://www.callweaver.org for more information about
 * the CallWeaver project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

/*!\file
 *
 * \brief Headerless G.726 (16/24/32/40kbps) data format for CallWeaver.
 * 
 * File name extensions:
 * \arg 40 kbps: g726-40
 * \arg 32 kbps: g726-32
 * \arg 24 kbps: g726-24
 * \arg 16 kbps: g726-16
 * \ingroup formats
 */
#ifdef HAVE_CONFIG_H
#include "confdefs.h"
#endif
 
#include <unistd.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdlib.h>
#include <sys/time.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>

#include "callweaver.h"

CALLWEAVER_FILE_VERSION("$HeadURL: https://svn.callweaver.org/callweaver/branches/rel/1.2/formats/format_g726.c $", "$Revision: 4723 $")

#include "callweaver/lock.h"
#include "callweaver/options.h"
#include "callweaver/channel.h"
#include "callweaver/file.h"
#include "callweaver/logger.h"
#include "callweaver/sched.h"
#include "callweaver/module.h"

#define RATE_40 0
#define RATE_32 1
#define RATE_24 2
#define RATE_16 3

/* We can only read/write chunks of FRAME_TIME ms G.726 data */
#define    FRAME_TIME    10    /* 10 ms size */

/* Frame sizes in bytes */
static int frame_size[4] =
{ 
    FRAME_TIME * 5,
    FRAME_TIME * 4,
    FRAME_TIME * 3,
    FRAME_TIME * 2
};

struct cw_filestream
{
    /* Do not place anything before "reserved" */
    void *reserved[CW_RESERVED_POINTERS];
    /* This is what a filestream means to us */
    FILE *f;                             /* Open file descriptor */
    int rate;                            /* RATE_* defines */
    struct cw_frame fr;                /* Frame information */
    char waste[CW_FRIENDLY_OFFSET];    /* Buffer for sending frames, etc */
    char empty;                            /* Empty character */
    unsigned char g726[FRAME_TIME * 5];    /* G.726 encoded voice */
};

CW_MUTEX_DEFINE_STATIC(g726_lock);
static int glistcnt = 0;

static char *desc = "Raw G.726 (16/24/32/40kbps) data";
static char *name40 = "g726-40";
static char *name32 = "g726-32";
static char *name24 = "g726-24";
static char *name16 = "g726-16";
static char *exts40 = "g726-40";
static char *exts32 = "g726-32";
static char *exts24 = "g726-24";
static char *exts16 = "g726-16";

/*
 * Rate dependant format functions (open, rewrite)
 */
static struct cw_filestream *g726_40_open(FILE *f)
{
    /* We don't have any header to read or anything really, but
       if we did, it would go here.  We also might want to check
       and be sure it's a valid file.  */
    struct cw_filestream *tmp;
    if ((tmp = malloc(sizeof(struct cw_filestream)))) {
        memset(tmp, 0, sizeof(struct cw_filestream));
        if (cw_mutex_lock(&g726_lock)) {
            cw_log(LOG_WARNING, "Unable to lock g726 list.\n");
            free(tmp);
            return NULL;
        }
        tmp->f = f;
        tmp->rate = RATE_40;
        cw_fr_init_ex(&tmp->fr, CW_FRAME_VOICE, CW_FORMAT_G726, name40);
        tmp->fr.data = tmp->g726;
        /* datalen will vary for each frame */
        glistcnt++;
        if (option_debug)
            cw_log(LOG_DEBUG, "Created filestream G.726-%dk.\n", 
                                    40 - tmp->rate * 8);
        cw_mutex_unlock(&g726_lock);
        cw_update_use_count();
    }
    return tmp;
}

static struct cw_filestream *g726_32_open(FILE *f)
{
    /* We don't have any header to read or anything really, but
       if we did, it would go here.  We also might want to check
       and be sure it's a valid file.  */
    struct cw_filestream *tmp;
    if ((tmp = malloc(sizeof(struct cw_filestream)))) {
        memset(tmp, 0, sizeof(struct cw_filestream));
        if (cw_mutex_lock(&g726_lock)) {
            cw_log(LOG_WARNING, "Unable to lock g726 list.\n");
            free(tmp);
            return NULL;
        }
        tmp->f = f;
        tmp->rate = RATE_32;
        cw_fr_init_ex(&tmp->fr, CW_FRAME_VOICE, CW_FORMAT_G726, name32);
        tmp->fr.data = tmp->g726;
        /* datalen will vary for each frame */
        glistcnt++;
        if (option_debug)
            cw_log(LOG_DEBUG, "Created filestream G.726-%dk.\n", 
                                    40 - tmp->rate * 8);
        cw_mutex_unlock(&g726_lock);
        cw_update_use_count();
    }
    return tmp;
}

static struct cw_filestream *g726_24_open(FILE *f)
{
    /* We don't have any header to read or anything really, but
       if we did, it would go here.  We also might want to check
       and be sure it's a valid file.  */
    struct cw_filestream *tmp;
    if ((tmp = malloc(sizeof(struct cw_filestream)))) {
        memset(tmp, 0, sizeof(struct cw_filestream));
        if (cw_mutex_lock(&g726_lock)) {
            cw_log(LOG_WARNING, "Unable to lock g726 list.\n");
            free(tmp);
            return NULL;
        }
        tmp->f = f;
        tmp->rate = RATE_24;
        cw_fr_init_ex(&tmp->fr, CW_FRAME_VOICE, CW_FORMAT_G726, name24);
        tmp->fr.data = tmp->g726;
        /* datalen will vary for each frame */
        glistcnt++;
        if (option_debug)
            cw_log(LOG_DEBUG, "Created filestream G.726-%dk.\n", 
                                    40 - tmp->rate * 8);
        cw_mutex_unlock(&g726_lock);
        cw_update_use_count();
    }
    return tmp;
}

static struct cw_filestream *g726_16_open(FILE *f)
{
    /* We don't have any header to read or anything really, but
       if we did, it would go here.  We also might want to check
       and be sure it's a valid file.  */
    struct cw_filestream *tmp;
    if ((tmp = malloc(sizeof(struct cw_filestream)))) {
        memset(tmp, 0, sizeof(struct cw_filestream));
        if (cw_mutex_lock(&g726_lock)) {
            cw_log(LOG_WARNING, "Unable to lock g726 list.\n");
            free(tmp);
            return NULL;
        }
        tmp->f = f;
        cw_fr_init_ex(&tmp->fr, CW_FRAME_VOICE, CW_FORMAT_G726, name16);
        tmp->rate = RATE_16;
        tmp->fr.data = tmp->g726;
        /* datalen will vary for each frame */
        glistcnt++;
        if (option_debug)
            cw_log(LOG_DEBUG, "Created filestream G.726-%dk.\n", 
                                    40 - tmp->rate * 8);
        cw_mutex_unlock(&g726_lock);
        cw_update_use_count();
    }
    return tmp;
}

static struct cw_filestream *g726_40_rewrite(FILE *f, const char *comment)
{
    /* We don't have any header to read or anything really, but
       if we did, it would go here.  We also might want to check
       and be sure it's a valid file.  */
    struct cw_filestream *tmp;
    if ((tmp = malloc(sizeof(struct cw_filestream)))) {
        memset(tmp, 0, sizeof(struct cw_filestream));
        if (cw_mutex_lock(&g726_lock)) {
            cw_log(LOG_WARNING, "Unable to lock g726 list.\n");
            free(tmp);
            return NULL;
        }
        tmp->f = f;
        tmp->rate = RATE_40;
        glistcnt++;
        if (option_debug)
            cw_log(LOG_DEBUG, "Created filestream G.726-%dk.\n", 
                                    40 - tmp->rate * 8);
        cw_mutex_unlock(&g726_lock);
        cw_update_use_count();
    } else
        cw_log(LOG_WARNING, "Out of memory\n");
    return tmp;
}

static struct cw_filestream *g726_32_rewrite(FILE *f, const char *comment)
{
    /* We don't have any header to read or anything really, but
       if we did, it would go here.  We also might want to check
       and be sure it's a valid file.  */
    struct cw_filestream *tmp;
    if ((tmp = malloc(sizeof(struct cw_filestream)))) {
        memset(tmp, 0, sizeof(struct cw_filestream));
        if (cw_mutex_lock(&g726_lock)) {
            cw_log(LOG_WARNING, "Unable to lock g726 list.\n");
            free(tmp);
            return NULL;
        }
        tmp->f = f;
        tmp->rate = RATE_32;
        glistcnt++;
        if (option_debug)
            cw_log(LOG_DEBUG, "Created filestream G.726-%dk.\n", 
                                    40 - tmp->rate * 8);
        cw_mutex_unlock(&g726_lock);
        cw_update_use_count();
    } else
        cw_log(LOG_WARNING, "Out of memory\n");
    return tmp;
}

static struct cw_filestream *g726_24_rewrite(FILE *f, const char *comment)
{
    /* We don't have any header to read or anything really, but
       if we did, it would go here.  We also might want to check
       and be sure it's a valid file.  */
    struct cw_filestream *tmp;
    if ((tmp = malloc(sizeof(struct cw_filestream)))) {
        memset(tmp, 0, sizeof(struct cw_filestream));
        if (cw_mutex_lock(&g726_lock)) {
            cw_log(LOG_WARNING, "Unable to lock g726 list.\n");
            free(tmp);
            return NULL;
        }
        tmp->f = f;
        tmp->rate = RATE_24;
        glistcnt++;
        if (option_debug)
            cw_log(LOG_DEBUG, "Created filestream G.726-%dk.\n", 
                                    40 - tmp->rate * 8);
        cw_mutex_unlock(&g726_lock);
        cw_update_use_count();
    } else
        cw_log(LOG_WARNING, "Out of memory\n");
    return tmp;
}

static struct cw_filestream *g726_16_rewrite(FILE *f, const char *comment)
{
    /* We don't have any header to read or anything really, but
       if we did, it would go here.  We also might want to check
       and be sure it's a valid file.  */
    struct cw_filestream *tmp;
    if ((tmp = malloc(sizeof(struct cw_filestream)))) {
        memset(tmp, 0, sizeof(struct cw_filestream));
        if (cw_mutex_lock(&g726_lock)) {
            cw_log(LOG_WARNING, "Unable to lock g726 list.\n");
            free(tmp);
            return NULL;
        }
        tmp->f = f;
        tmp->rate = RATE_16;
        glistcnt++;
        if (option_debug)
            cw_log(LOG_DEBUG, "Created filestream G.726-%dk.\n", 
                                    40 - tmp->rate * 8);
        cw_mutex_unlock(&g726_lock);
        cw_update_use_count();
    } else
        cw_log(LOG_WARNING, "Out of memory\n");
    return tmp;
}

/*
 * Rate independent format functions (close, read, write)
 */
static void g726_close(struct cw_filestream *s)
{
    if (cw_mutex_lock(&g726_lock)) {
        cw_log(LOG_WARNING, "Unable to lock g726 list.\n");
        return;
    }
    glistcnt--;
    if (option_debug)
        cw_log(LOG_DEBUG, "Closed filestream G.726-%dk.\n", 40 - s->rate * 8);
    cw_mutex_unlock(&g726_lock);
    cw_update_use_count();
    fclose(s->f);
    free(s);
    s = NULL;
}

static struct cw_frame *g726_read(struct cw_filestream *s, int *whennext)
{
    int res;

    /* Send a frame from the file to the appropriate channel */
    cw_fr_init_ex(&s->fr, CW_FRAME_VOICE, CW_FORMAT_G726, NULL);
    s->fr.offset = CW_FRIENDLY_OFFSET;
    s->fr.samples = 8*FRAME_TIME;
    s->fr.datalen = frame_size[s->rate];
    s->fr.data = s->g726;
    if ((res = fread(s->g726, 1, s->fr.datalen, s->f)) != s->fr.datalen)
    {
        if (res)
            cw_log(LOG_WARNING, "Short read (%d) (%s)!\n", res, strerror(errno));
        return NULL;
    }
    *whennext = s->fr.samples;
    return &s->fr;
}

static int g726_write(struct cw_filestream *fs, struct cw_frame *f)
{
    int res;
    if (f->frametype != CW_FRAME_VOICE) {
        cw_log(LOG_WARNING, "Asked to write non-voice frame!\n");
        return -1;
    }
    if (f->subclass != CW_FORMAT_G726) {
        cw_log(LOG_WARNING, "Asked to write non-G726 frame (%d)!\n", 
                        f->subclass);
        return -1;
    }
    if (f->datalen % frame_size[fs->rate]) {
        cw_log(LOG_WARNING, "Invalid data length %d, should be multiple of %d\n", 
                        f->datalen, frame_size[fs->rate]);
        return -1;
    }
    if ((res = fwrite(f->data, 1, f->datalen, fs->f)) != f->datalen) {
            cw_log(LOG_WARNING, "Bad write (%d/%d): %s\n", 
                            res, frame_size[fs->rate], strerror(errno));
            return -1;
    }
    return 0;
}

static char *g726_getcomment(struct cw_filestream *s)
{
    return NULL;
}

static int g726_seek(struct cw_filestream *fs, long sample_offset, int whence)
{
    return -1;
}

static int g726_trunc(struct cw_filestream *fs)
{
    return -1;
}

static long g726_tell(struct cw_filestream *fs)
{
    return -1;
}

/*
 * Module interface (load_module, unload_module, usecount, description, key)
 */
int load_module(void)
{
    int res;

    res = cw_format_register(name40, exts40, CW_FORMAT_G726,
                               g726_40_open,
                               g726_40_rewrite,
                               g726_write,
                               g726_seek,
                               g726_trunc,
                               g726_tell,
                               g726_read,
                               g726_close,
                               g726_getcomment);
    if (res)
    {
        cw_log(LOG_WARNING, "Failed to register format %s.\n", name40);
        return(-1);
    }
    res = cw_format_register(name32, exts32, CW_FORMAT_G726,
                               g726_32_open,
                               g726_32_rewrite,
                               g726_write,
                               g726_seek,
                               g726_trunc,
                               g726_tell,
                               g726_read,
                               g726_close,
                               g726_getcomment);
    if (res)
    {
        cw_log(LOG_WARNING, "Failed to register format %s.\n", name32);
        return(-1);
    }
    res = cw_format_register(name24, exts24, CW_FORMAT_G726,
                               g726_24_open,
                               g726_24_rewrite,
                               g726_write,
                               g726_seek,
                               g726_trunc,
                               g726_tell,
                               g726_read,
                               g726_close,
                               g726_getcomment);
    if (res)
    {
        cw_log(LOG_WARNING, "Failed to register format %s.\n", name24);
        return(-1);
    }
    res = cw_format_register(name16, exts16, CW_FORMAT_G726,
                                g726_16_open,
                                g726_16_rewrite,
                                g726_write,
                                g726_seek,
                                g726_trunc,
                                g726_tell,
                                g726_read,
                                g726_close,
                                g726_getcomment);
    if (res) {
        cw_log(LOG_WARNING, "Failed to register format %s.\n", name16);
        return(-1);
    }
    return(0);
}

int unload_module(void)
{
    int res;

    res = cw_format_unregister(name16);
    if (res) {
        cw_log(LOG_WARNING, "Failed to unregister format %s.\n", name16);
        return(-1);
    }
    res = cw_format_unregister(name24);
    if (res) {
        cw_log(LOG_WARNING, "Failed to unregister format %s.\n", name24);
        return(-1);
    }
    res = cw_format_unregister(name32);
    if (res) {
        cw_log(LOG_WARNING, "Failed to unregister format %s.\n", name32);
        return(-1);
    }
    res = cw_format_unregister(name40);
    if (res) {
        cw_log(LOG_WARNING, "Failed to unregister format %s.\n", name40);
        return(-1);
    }
    return(0);
}    

int usecount(void)
{
    return glistcnt;
}

char *description(void)
{
    return desc;
}
